from uuid import uuid4

import httpx
import pytest
from fastapi import status
from mesh_client import MeshClient
from requests.models import HTTPError

from mesh_sandbox.tests.docker_tests import (
    _CANNED_MAILBOX1,
    _CANNED_MAILBOX2,
    _PASSWORD,
    _SHARED_KEY,
)


def _get_inbox_count(base_uri: str, recipient_mailbox_id: str):
    with MeshClient(
        url=base_uri, mailbox=recipient_mailbox_id, password=_PASSWORD, shared_key=_SHARED_KEY
    ) as recipient:
        message_ids = recipient.list_messages()
        return len(message_ids)


def _send_message_and_assert_inbox_count_is_one(
    base_uri: str, recipient_mailbox_id: str, workflow_id: str, payload: bytes
):
    sender_mailbox_id = _CANNED_MAILBOX1

    with MeshClient(
        url=base_uri, mailbox=sender_mailbox_id, password=_PASSWORD, shared_key=_SHARED_KEY, max_chunk_size=100
    ) as sender:

        message_id = sender.send_message(recipient_mailbox_id, payload, workflow_id=workflow_id)
        assert message_id
        assert _get_inbox_count(base_uri, _CANNED_MAILBOX2) == 1
        return message_id


def test_app_health(base_uri: str):

    with httpx.Client(base_url=base_uri) as client:
        res = client.get("/health")
        assert res.status_code == status.HTTP_200_OK


def test_handshake(base_uri: str):

    with MeshClient(url=base_uri, mailbox=_CANNED_MAILBOX1, password=_PASSWORD, shared_key=_SHARED_KEY) as client:
        res = client.handshake()
        assert res == b"hello"


def test_handshake_bad_password(base_uri: str):

    with MeshClient(url=base_uri, mailbox=_CANNED_MAILBOX1, password="BAD", shared_key=_SHARED_KEY) as client:

        with pytest.raises(HTTPError) as err:
            client.handshake()
        assert err.value.response.status_code == status.HTTP_400_BAD_REQUEST


def test_send_receive_chunked_message(base_uri: str):

    recipient_mailbox_id = _CANNED_MAILBOX2
    workflow_id = uuid4().hex
    sent_payload = b"a" * 1000
    message_id = _send_message_and_assert_inbox_count_is_one(base_uri, recipient_mailbox_id, workflow_id, sent_payload)

    with MeshClient(
        url=base_uri, mailbox=recipient_mailbox_id, password=_PASSWORD, shared_key=_SHARED_KEY
    ) as recipient:
        message_ids = recipient.list_messages()
        assert message_ids == [message_id]
        message = recipient.retrieve_message(message_id)
        assert message.workflow_id == workflow_id  # pylint: disable=no-member
        received_payload = message.read()
        assert received_payload == sent_payload

        message.acknowledge()
        message_ids = recipient.list_messages()
        assert message_ids == []
